import sisu_email.create


def encode_multipart_message(message):
    """Encode multipart message_id as urlsafe base64 string

    :param message: email.mime.multipart.MIMEMultipart
    :return: dict, {'raw': base64_string_of_message}
    """
    return {
        'raw': sisu_email.create.encode_multipart_message(message)
    }




def download_message(resource, user_id, message_id):
    """Download a message by id

    :param resource: Gmail API Resource
    :param user_id: message_id owner Gmail API userId
    :param message_id: str, Gmail message id
    :return: dict of message contents
    """
    return resource.users().messages().get(
        userId=user_id,
        id=message_id
    ).execute()



def send_message(resource, user_id, message):
    """Send email via Gmail API

    :param resource: Gmail API Resource
    :param user_id: sender Gmail API userId
    :param message: email.mime.multipart.MIMEMultipart
    :return: Gmail API response
    """
    message = resource.users().messages().send(
        userId=user_id,
        body=message
    ).execute()
    return message




def get_profile(resource, user_id):
    """Fetch profile data for a given Gmail userId

    :param resource: Gmail API resource
    :param user_id: str, Gmail API userId
    :return: dict, Gmail API profile
    """
    return resource.users().getProfile(
        userId=user_id,
    ).execute()



import os
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build


def start_auth_flow(app_creds_path, token_path, scopes):
    """Start Google auth flow using app_creds at path_to_app_creds

    :param app_creds_path: path_to_app_creds to app_creds.json file
    :param token_path: path_to_app_creds to store resulting auth app_creds
    :param scopes: Gmail API permission scope to request
    :return: token_path to the app_creds
    """
    creds = None
    if os.path.exists(token_path):
        creds = Credentials.from_authorized_user_file(
            token_path,
            scopes
        )

    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(app_creds_path, scopes)
            creds = flow.run_local_server(port=0)

        with open(token_path, 'w') as token_file:
            token_file.write(creds.to_json())

    return token_path


def authorize_resource(credentials):
    """Create an authorized Gmail API resource

    :param credentials: google.oauth2.app_creds.Credentials
    :return: googleapiclient.discovery.Resource
    """
    return build('gmail', 'v1', credentials=credentials)


def creds_from_json(token_json):
    """Create google.oauth2.app_creds from token_json

    :param token_json: dict: Gmail API user_token json loaded as dict
    :return: google.oauth2.app_creds.Credentials
    """
    return Credentials.from_authorized_user_info(token_json)


def refresh_token(credentials):
    """Call api with an empty Request to refresh token_path

    :param credentials: google.oauth2.credentials.Credentials
    :return: credentials: google.oauth2.credentials.Credentials
    """
    credentials.refresh(Request())
    return credentials



import inspect


def get_labels(resource, user_id):
    """Get a list of message labels for user_id
    
    :param resource: Gmail API Resource
    :param user_id: sender Gmail API userId
    :return: Gmail API response
    """
    response = resource.users().labels().list(
        userId=user_id
    ).execute()
    return response


def add_labels(resource, user_id, message_id, labels):
    """Apply a list of labels to a Gmail message

    :param resource: Gmail API resource
    :param user_id: str, Gmail API userId
    :param message_id: str, Gmail message id
    :param labels: list, str, labels to apply to message
    :return: Gmail API response
    """
    if not type(labels) is list:
        raise ValueError(
            f'{inspect.stack()[0][3]} requires list for labels arg'
        )
    response = resource.users().messages().modify(
        userId=user_id,
        id=message_id,
        body={'addLabelIds': labels}
    ).execute()
    return response


def get_label_by_name(resource, user_id, name):
    """Case insensitive search for a label by name

    :param resource: Gmail API Resource
    :param user_id: str, Gmail API userId
    :param name: str, name of label to try to get
    :return: Gmail label object or None
    """
    labels = get_labels(resource, user_id)
    for label in labels['labels']:
        if label['name'].lower() == name.lower():
            return label
    else:
        return None


def remove_labels(resource, user_id, message_id, labels):
    """Remove a list of labels from a Gmail message

    :param resource: Gmail API resource
    :param user_id: str, Gmail API userId
    :param message_id: str, Gmail message id
    :param labels: list, str, list of label ids to remove
    :return: Gmail API response
    """
    if not type(labels) is list:
        raise ValueError(
            f'{inspect.stack()[0][3]} requires list for labels arg'
        )
    response = resource.users().messages().modify(
        userId=user_id,
        id=message_id,
        body={'removeLabelIds': labels}
    ).execute()
    return response



def delete_message(resource, user_id, message_id):
    """Send email via Gmail API

    :param resource: Gmail API Resource
    :param user_id: message_id owner Gmail API userId
    :param message_id: Gmail API message_id id
    :return: Gmail API response
    """
    response = resource.users().messages().delete(
        userId=user_id,
        id=message_id
    ).execute()
    return response


def batch_delete(resource, user_id, message_ids):
    """Batch delete a list of Gmail messages

    :param resource: Gmail API Resource
    :param user_id: message_id owner Gmail API userId
    :param message_ids: list, dict, Gmail message_id objects
    :return: dict, Gmail API response
    """
    payload = {"ids": [msg['id'] for msg in message_ids]}
    response = resource.users().messages().batchDelete(
        userId=user_id,
        body=payload
    ).execute()
    return response



QUERIES = {
    'FROM_ADDRESS': 'from:'
}


class NoNextPageToken(KeyError):
    pass


def next_page(resource, user_id, query, response, max_results=100):
    """Get next page of search results

    :param resource: Gmail API resource
    :param user_id: str, Gmail API userId
    :param query: str, Gmail API search query
    :param response: dict, Gmail API search query response
    :param max_results: int, number of results per request
    :return: list, dicts of gmail message_id message_ids and thread message_ids
    """
    if not type(response) is dict:
        raise TypeError(
            'next_search_page requires dict as response param'
        )
    elif 'nextPageToken' not in response:
        raise NoNextPageToken(
            'next_search_page requires dict with nextPageToken key'
        )

    page_token = response['nextPageToken']
    return resource.users().messages().list(
        userId=user_id,
        q=query,
        pageToken=page_token,
        maxResults=max_results
    ).execute()


def search(resource, user_id, query, max_results=100):
    """Return search results response for a Gmail API query

    :param resource: Gmail API resource
    :param user_id: str, Gmail API userId
    :param query: str, Gmail API search query
    :param max_results: int, number of results per request
    :return: Gmail API search response
    """
    return resource.users().messages().list(
        userId=user_id,
        q=query,
        maxResults=max_results
    ).execute()


def iter_messages(resource, user_id, query):
    """Generator to return search results

    :param resource: Gmail API resource
    :param user_id: str, Gmail API userId
    :param query: str, Gmail API search query
    :return: dict, gmail message_id message_ids and thread message_ids
    """
    response = resource.users().messages().list(
        userId=user_id,
        q=query
    ).execute()

    if 'messages' in response:
        part = response['messages']
        for index, result in enumerate(part, start=1):
            yield result
            if len(part) == index and 'nextPageToken' in response:
                next_page(resource, user_id, response, query)


def search_by_address(resource, user_id, address):
    """Search gmail for messages by sender email address

    :param resource: Gmail API Resource
    :param user_id: str, Gmail API userId
    :param address: str, Gmail address
    :return: list, dict of Gmail message ids and thread ids:
    """
    return search(
        resource,
        user_id,
        f"{QUERIES['FROM_ADDRESS']}{address}"
    )
